/* MausConnect: Nachrichtenformatskonverter fuer das MausNet
 * Copyright (C) 1994 1995 Uwe Ohse
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * nntp.c: Routinen fr nntp 
 */

#include <stdio.h>

#include "mausconsys.h"

#include <conflib.h>
#include <syslog.h>
#include <assert.h>

#include "mauscon.h"
#include "mausconlib.h"
#include "metamsg.h"
#include "rfc.h"
#include "nntp.h"
#include "msg_io.h"

	/* nntp-Zeilenende */
#define CRLF "\r\n"

struct nntp_private
{
	FILE *in;
	FILE *out;
	char *host;
	char *fname;
	char nntprepbuffer[1024];
	BOOLEAN is_suck;
	BOOLEAN may_close;
	buf_t mbuf;
};

static mio_retcode_t nntp_export (mio_t * chan, metamsg_t *);
#ifdef WANT_NNTP
static mio_retcode_t nntp_open (const char *host, const char *mode,
								mio_t ** neu, char *errbuf);
#endif
static mio_retcode_t nntp_close (mio_t * chan);
static mio_retcode_t suck_open (const char *host, const char *mode,
								mio_t ** neu, char *errbuf);
static mio_retcode_t suck_import (mio_t * chan, metamsg_t *);
#ifdef WANT_NNTP
static char *format_reply (struct nntp_private *);
static int nntp_get_reply (struct nntp_private *priv);
#endif

#ifdef WANT_NNTP
mio_base_t mio_nntp =
{
	"nntp",

	nntp_open,
	NULL,						/* kein Import mglich */
	nntp_export,
	nntp_close,

	0
};
#endif

mio_base_t mio_suck =
{
	"suck",

	suck_open,
	suck_import,				/* eingeschrnkt mglich */
	nntp_export,
	nntp_close,

	0
};

mio_retcode_t 
nntp_close (mio_t * chan)
{
	struct nntp_private *priv = (struct nntp_private *) chan->priv_data;

#ifdef WANT_NNTP
	if (!priv->is_suck)
	{
		fputs ("QUIT" CRLF, priv->out);
		(void) nntp_get_reply (priv);
	}
#endif
	if (priv->may_close)
	{
		if (priv->out)
			fclose (priv->out);
		if (priv->in)
		fclose (priv->in);
	}
	free (priv->host);
	free (priv->fname);
	free (priv->mbuf.start);
	free (priv);
	chan->priv_data = NULL;
	mio_free (chan);
	return MIO_OK;
}


static mio_retcode_t
suck_open (const char *fname, const char *mode, mio_t ** ret_handle, char *errbuf)
{
	struct nntp_private *priv;
	mio_t *neu;
	BOOLEAN output;

	if (*mode == 'w')
		output = TRUE;
	else if (strcmp (mode, "r")==0)
		output = FALSE;
	else
	{
		sprintf (errbuf, "illegal mode %s for suck_open", mode);
		log (LOG_ERR, "Illegal mode %s for suck_open", mode);
		return MIO_ERROR;
	}

	neu = mio_alloc (&mio_suck);
	priv = malloc (sizeof (struct nntp_private));
	if (priv)
		priv->fname = strdup(fname);

	if (!neu || !priv || !priv->fname)
	{
		strcpy (errbuf, _("out of memory"));
		log (LOG_ERR, _("out of memory"));
		if (priv)
			free(priv->fname);
		free (priv);
		free (neu);
		return MIO_RETRY;
	}
	priv->is_suck = TRUE;
	priv->may_close = TRUE;

	if (output)
	{
		priv->in = NULL;
		if (0 == strcmp (fname, "-"))
		{
			priv->out = stdout;
			priv->may_close = FALSE;
		}
		else
		{
			priv->out = fopen (fname, mode);
			if (!priv->out)
			{
				int errno1 = errno;

				free (priv->fname);
				free (priv);
				free (neu);
				sprintf (errbuf, _("cannot open %s: %s"), fname, strerror (errno1));
				log (LOG_ERR, _("cannot open %s: %s"), fname, strerror (errno1));
				return MIO_RETRY;
			}
			setvbuf (priv->out, malloc (16384), _IONBF, 16384);
		}
	}
	else
	{
		priv->out = NULL;
		if (0 == strcmp (fname, "-"))
		{
			priv->in = stdin;
			priv->may_close = FALSE;
		}
		else
		{
			priv->in = fopen (fname, mode);
			if (!priv->in)
			{
				int errno1 = errno;

				free (priv->fname);
				free (priv);
				free (neu);
				sprintf (errbuf, _("cannot open %s: %s"), fname, strerror (errno1));
				log (LOG_ERR, _("cannot open %s: %s"), fname, strerror (errno1));
				return MIO_RETRY;
			}
			setvbuf (priv->in, malloc (16384), _IOLBF, 16384);
		}
	}

#define STARTSIZE (64*1024L)
	priv->mbuf.start=XMALLOC(STARTSIZE);
	priv->mbuf.next=priv->mbuf.start;
	priv->mbuf.size=STARTSIZE;

	neu->priv_data = priv;
	*ret_handle = neu;
	return MIO_OK;
}

#ifdef WANT_NNTP
static mio_retcode_t
nntp_open (const char *host, const char *mode, mio_t ** ret_handle, char *errbuf)
{
	int ret;
	struct nntp_private *priv;
	mio_t *neu;
	char *errtext = NULL;

	neu = mio_alloc (&mio_nntp);
	priv = malloc (sizeof (struct nntp_private));

	if (priv)
		priv->host = strdup (host);
	if (!neu || !priv || !priv->host)
	{
		strcpy (errbuf, _("out of memory"));
		log (LOG_ERR, _("out of memory"));
		if (priv)
			free (priv->host);
		free (priv);
		free (neu);
		return MIO_RETRY;
	}

	if (!fopen_service ("nntp", host, &priv->in, &priv->out, &errtext))
	{
		log (LOG_ERR, "%s", errtext ? errtext : _("out of memory"));
		sprintf (errbuf, "%s", errtext ? errtext : _("out of memory"));
		return MIO_RETRY;
	}

	setvbuf (priv->in, malloc (16384), _IOLBF, 16384);
	setvbuf (priv->out, NULL, _IONBF, 0);
	priv->nntprepbuffer[0] = 0;

	ret = nntp_get_reply (priv);	/* Connection establishment reply */
	if (ret > 299 || ret == 0)
	{
	   /* Tja, keine Verbindung ... */
		if (ret)
			log (LOG_WARNING, _("Connection not established: %s"), priv->nntprepbuffer);
		else
			log (LOG_WARNING, _("Connection timed out"));
		fclose (priv->out);
		fclose (priv->out);
		return MIO_RETRY;
	}
	priv->mbuf.start=NULL;
	priv->mbuf.next=NULL;
	priv->mbuf.size=0;
	priv->fname=NULL;
	priv->may_close=TRUE;
	priv->is_suck=FALSE;

	neu->priv_data = priv;
	*ret_handle = neu;
	return MIO_OK;
}
#endif

#ifdef WANT_NNTP
static int
nntp_get_reply (struct nntp_private *priv)
{
	fd_set fds;
	struct timeval timeout;
	int selcode;

	FD_ZERO (&fds);
	FD_SET (fileno (priv->in), &fds);

	timeout.tv_sec = 10 * 60;	/* seconds */
	timeout.tv_usec = 0;

	priv->nntprepbuffer[0] = '\0';

   /* `select' returns 0 if timeout, 1 if input available, -1 if error. */
	do
	{
		selcode = select (FD_SETSIZE, &fds, NULL, NULL, &timeout);
	}
	while (selcode == -1);		/* einschlagende Signale oder so */
	if (select (FD_SETSIZE, &fds, NULL, NULL, &timeout) != 1)
		return 000;				/* Timeout signalisieren */

   /* nntp kennt keine Fortsetzungen */
	fgets (priv->nntprepbuffer, sizeof (priv->nntprepbuffer), priv->in);
	return (int) strtol (priv->nntprepbuffer, NULL, 10);
}
#endif

#ifdef WANT_NNTP
static
char *
format_reply (struct nntp_private *priv)
{
	char *p = priv->nntprepbuffer;
	char *p2;

	while (isspace (*p) || isdigit (*p))
		p++;
	p2 = strlen (p) + p - 1;
	while (p2 > p && (*p2 == '\r' || *p2 == '\n'))
		*p2-- = '\0';
	return p;
}
#endif

static mio_retcode_t
suck_import (mio_t * chan, metamsg_t * meta)
{
#define BLOCK 65536
#define REST (mbuf->size - (mbuf->next-mbuf->start))
	struct nntp_private *priv;
	buf_t *mbuf;
	BOOLEAN in_header=TRUE;
	BOOLEAN normal_end=FALSE;

	priv = (struct nntp_private *) chan->priv_data;
	mbuf=&priv->mbuf;
	mbuf->next=mbuf->start;

	while (fgets(mbuf->next, REST, priv->in))
	{
		size_t l;

		if (0==strcmp(mbuf->next,".\n")
			|| 0==strcmp(mbuf->next,".\r\n"))
		{
			*mbuf->next=0;
			normal_end=TRUE;
			break; /* Ende der Nachricht */
		}
		if (*mbuf->next=='.')
		{
			/* Escape/Dot beseitigen */
			l=strlen(mbuf->next);
			memmove(mbuf->next,mbuf->next+1,l);
			l--;
		}
		else
			l=strlen(mbuf->next);
		/* \r\n -> \n komprimieren. Ntig? */
		if (l>1 && mbuf->next[l-2]=='\r')
		{
			mbuf->next[l-2]='\n';
			mbuf->next[l-1]='\0';
			l--;
		}

		if (in_header)
		{
			if (*mbuf->next=='\n')
			{
				in_header=FALSE;
			}
			else if ((*mbuf->next==' ' || *mbuf->next=='\t')
				&& mbuf->start!=mbuf->next)
			{
				/* Fortsetzungszeile */
				char *p;
				p=mbuf->next;
				while (isspace(*p))
					p++;
				if (*p)
					memmove(mbuf->next-1,p,strlen(p)+1);
				else
					l=0; /* Leere Zeile *kommt* *selten* vor */
			}
		}
		mbuf->next+=l;

		if (REST < BLOCK)
		{
			off_t offset;
			offset=mbuf->next-mbuf->start;
			mbuf->start=XREALLOC(mbuf->start,mbuf->size+BLOCK+1);
			mbuf->next=mbuf->start+offset;
			mbuf->size+=BLOCK+1;
		}
	}
	if (mbuf->next==mbuf->start)
		/* wenn nicht gelesen wurde sind wir auch nicht im Header */
		in_header=FALSE;

	if (feof(priv->in) && mbuf->next==mbuf->start)
	{
		return MIO_EOF;
		normal_end=TRUE;
	}

	/* Batches ohne abschlieenden Punkt erlauben */
	if (feof(priv->in) && !in_header)
		normal_end=TRUE;

	if (!normal_end)
	{
		/* Diese Nachricht ist *Schrott* */
		int errno1=errno;
		mio_set_errtext(chan, _("cannot read from %s: %s"),
			priv->fname ? priv->fname : "???", strerror(errno1));
		log(LOG_ERR,_("cannot read from %s: %s"),
			priv->fname ? priv->fname : "???", strerror(errno1));
		return MIO_EOF;
	}

	if (in_header)
	{
		/* Diese Nachricht ist *Schrott* */
		mio_set_errtext(chan,_("message ended in header: %s"),
			priv->fname ? priv->fname : "???");
		log(LOG_ERR,_("message ended in header: %s"),
			priv->fname ? priv->fname : "???");
		return MIO_ERROR;
	}

	if (rfc_to_meta(mbuf,meta))
	{
		/* Fehler beim Umwandeln einer Nachricht. Nun gut, allzu
		 * schlimm ist das nicht, weiter mit der nchsten.
		 * Hufigste Ursache: Kaputter Header, unparsbare Adresse
		 * oder Datum.
		 */
		return MIO_ERROR;
	}
	return MIO_OK;
}

#define RETCODE(r) ((r==400) ? 2 : 1 )
static mio_retcode_t
nntp_export (mio_t * chan, metamsg_t * meta)
{
#ifdef WANT_NNTP
	int ret;
#endif
	struct nntp_private *priv;

	priv = (struct nntp_private *) chan->priv_data;

#ifdef WANT_NNTP
	if (!priv->is_suck)
	{
		fprintf (priv->out, "IHAVE <%s>" CRLF, meta->message_id);
		ret = nntp_get_reply (priv);	/* IHAVE reply */

		if (ret != NNTP_SENDIT_VAL)
		{
			char *p = format_reply (priv);

		   /* Aha? Warum das? */
			if (ret == NNTP_GOODBYE_VAL)	/* Verbindung zusammengebrochen */
			{
				msglog ("# %s: NNTP %s", meta->message_id, p);
				log (LOG_WARNING, _("connection to NNTP server lost: %s"), p);
				mio_set_errtext (chan, "%s", p);
				return MIO_ERROR;
			}
			if (ret == NNTP_HAVEIT_VAL)		/* typischerweise Dupe */
			{
				msglog ("! %s: NNTP %s", meta->message_id, p);
				return MIO_OK;
			}
			if (ret == NNTP_RESENDIT_VAL)
			{
				msglog ("! %s: NNTP %s", meta->message_id, p);
				mio_set_errtext (chan, "%s", p);
				return MIO_RETRY;
			}
			if (ret == 0)
			{
				msglog ("! %s: NNTP connection timed out", meta->message_id);
				mio_set_errtext (chan, _("connection timed out"));
				return MIO_RETRY;
			}
			if (ret == NNTP_AUTH_NEEDED_VAL) {
				/* INNs nnrpd sends this.We have no chance to send with
				 * NNTP, because we already are in nnrpd */
				msglog ("! %s: NNTP %s", meta->message_id, p);
				log (LOG_NOTICE, _("connection to NNTP server lost: %s"), p);
				mio_set_errtext (chan, "%s", p);
				return MIO_ERROR;
			}
			mio_set_errtext (chan, "%s", p);
			msglog ("! %s: NNTP %s", meta->message_id, p);
			return MIO_RETRY;		/* irgendein marginales oder greres Problem */
		}
	}
#endif

	rfc_print_message (meta, priv->out, CRLF, 1);
	fputs ("." CRLF, priv->out);
#ifdef WANT_NNTP
	if (!priv->is_suck)
	{
		ret = nntp_get_reply (priv);
		if (ret == 0)
		{
			msglog ("! %s: NNTP connection timed out", meta->message_id);
			mio_set_errtext (chan, _("connection timed out"));
			return MIO_RETRY;
		}
		else if (ret > 299)
		{
			char *p = format_reply (priv);

			mio_set_errtext (chan, "%s", p);
			if (ret == NNTP_GOODBYE_VAL)
			{
				log (LOG_WARNING, _("connection to NNTP server lost: %s"), p);
				msglog ("! %s: NNTP %s", meta->message_id, p);
				/* nach oben melden, da der Kanal nun unbenutzbar ist */
				return MIO_ERROR;
			}
			if (ret == NNTP_RESENDIT_VAL)
			{
				msglog ("! %s: NNTP %s", meta->message_id, p);
				return MIO_RETRY;
			}
			msglog ("! %s: NNTP %s", meta->message_id, p);
			return MIO_RETRY;		/* irgendein marginales oder greres Problem */
		}
		msglog ("  %s: NNTP ok", meta->message_id);
	}
	else
#endif
		msglog ("  %s: suck -> ok", meta->message_id);
	return MIO_OK;
}
#undef RETCODE


/* hier folgt all das, was mit dem eigentlichen Nachrichtentransport 
 * nichts zu tun hat, sondern nur dem Einholen von Informationen 
 * dient.
 */

#ifdef WANT_NNTP
static mio_t *inquire;

BOOLEAN
nntp_inquire_open (const char *host)
{
	char errbuf[1024];

	if (MIO_OK == nntp_open (host, "w", &inquire, errbuf))
	{
		return TRUE;
	}
	log (LOG_ERR, "nntp_open(%s): %s", host, errbuf);
	return FALSE;
}

void
nntp_inquire_close (void)
{
	if (inquire)
		inquire->fn->close (inquire);
}

static int
nntp_inquire_art_cmd (const char *cmd, const char *id, long artnum)
{
	struct nntp_private *priv = (struct nntp_private *) inquire->priv_data;

	if (id)
		fprintf (priv->out, "%s <%s>" CRLF, cmd, id);
	else
		fprintf (priv->out, "%s %ld" CRLF, cmd, artnum);
	return nntp_get_reply (priv);
}

int
nntp_inquire_stat (const char *id, long artnum)
{
	return nntp_inquire_art_cmd ("STAT", id, artnum);
}

int
nntp_inquire_head (const char *id, long artnum)
{
	return nntp_inquire_art_cmd ("HEAD", id, artnum);
}

BOOLEAN
nntp_inquire_get_list (char **buf, size_t * bufsize)
{
	char lbuf[16384];			/* eine Zeile :-) */
	char *wp;
	char *cp;
	int ende = 0;
	size_t l;
	struct nntp_private *priv;

	wp = *buf;

	priv = (struct nntp_private *) inquire->priv_data;

	while (!ende)
	{
		if (!fgets (lbuf, sizeof (lbuf), priv->in))
			ende = 1;
		else if (*lbuf == '.')
		{
		   /* Ende-Zeile? */
			if ((lbuf[1] == '\r' && lbuf[2] == '\n' && lbuf[3] == '\0')
				|| (lbuf[1] == '\n' && lbuf[2] == '\0'))
				ende = 1;
			else
				cp = lbuf + 1;
		}
		else
			cp = lbuf;
		if (!ende)
		{
			l = strlen (cp);
			if (!*buf || !*bufsize)
			{
				*buf = XMALLOC (16384);
				*bufsize = 16384;
				wp = *buf;
			}
			else if ((wp - *buf) + l > *bufsize - 1)
			{
				size_t offset;

				offset = wp - *buf;
				*bufsize += 16384;
				*buf = XREALLOC (*buf, *bufsize);
				wp = *buf + offset;
			}
			memcpy (wp, cp, l);
			wp += l;
		}
	}							/* while ! ende */
	*wp = 0;
	return TRUE;
}

char *
nntp_inquire_get_headerentry (const char *id, const char *kennung)
{
	int ret;
	char *p,
	 *q;
	static size_t hbufsize;
	static char *hbuf;
	static char last_id[512] = "";

	if (strcasecmp (id, last_id))
	{
		ret = nntp_inquire_head (id, 0);
		if (ret > 299)
			return NULL;

	   /* lesen */
		if (!nntp_inquire_get_list (&hbuf, &hbufsize))
			return NULL;
	   /* So - der ganze Header ist in hbuf ... */
	}

   /* Das sieht nach bertriebenem Aufwand aus, wo man doch auch mit
    * * XHDR an einzelne Zeilen kommen kann. Leider jedoch ffnen der
    * * Referenz-NNTP-Server und INN fr jeden Header erneut das File,
    * * und das frit Zeit ...
    */
	p = rfc_find_header (hbuf, kennung);
	if (!p)
		return NULL;
	while (1)
	{
		q = strchr (hbuf, '\n');
		if (!q)
			break;
		if (!isspace (q[1]))
		{
			if (q > hbuf && q[-1] == '\r')
				q[-1] = '\0';
			else
				*q = '\0';
			break;
		}
		if (q > hbuf && q[-1] == '\r')
			q[-1] = ' ';
		*q = ' ';
	}
	return p;
}

#endif /* WANT_NNTP */
