/* MausConnect: Nachrichtenformatskonverter fuer das MausNet
   Copyright (C) 1994 1995 Uwe Ohse

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   smtp.c: Routinen fr smtp (simple mail transfer protocol)
*/

#include "mausconsys.h"

#include <stdio.h>
#include <assert.h>
#include <syslog.h>

#include "mausconlib.h"
#include "mauscon.h"
#include "metamsg.h"
#include "rfc.h"
#include "msg_io.h"

#define CRLF "\r\n"

struct smtp_private 
{
	BOOLEAN batch;
	FILE *in;
	FILE *out;
	char *host;
	char *fname;
	char smtprepbuffer[1024];
};

static mio_retcode_t bsmtp_open(const char *fname, const char *mode,
	mio_t **neu, char *errbuf);
static mio_retcode_t smtp_close(mio_t *chan);
static mio_retcode_t smtp_export(mio_t *chan,metamsg_t *);
#if defined(WANT_SMTP)
static mio_retcode_t smtp_open(const char *host, const char *mode,
	mio_t **neu, char *errbuf);
static char *format_reply(struct smtp_private *);
static mio_retcode_t retcode_and_log(struct smtp_private *,int r, const char *);
static int smtp_get_reply (struct smtp_private *priv);
#endif


#if defined(WANT_SMTP)
mio_base_t mio_smtp=
{
	"smtp",

	smtp_open,
	NULL,         /* kein Import zur Zeit mglich */
	smtp_export,
	smtp_close,

	0
};
#endif

mio_base_t mio_batch_smtp=
{
	"bsmtp",

	bsmtp_open,
	NULL,
	smtp_export,
	smtp_close,

	0
};

#if defined(WANT_SMTP)
static char *
format_reply(struct smtp_private *priv)
{
	char *p=priv->smtprepbuffer;
	char *p2;
	while (isspace(*p) && isdigit(*p))
		p++;
	p2=strlen(p)+p-1;
	while (*p2=='\r' || *p2=='\n')
		*p2--='\0';
	return p;
}

static int
smtp_get_reply (struct smtp_private *priv)
{
	fd_set fds;
	struct timeval timeout;
	int cont = 1;

	FD_ZERO (&fds);
	FD_SET (fileno (priv->in), &fds);

	/* Initialize the timeout data structure. */
	/* RFC 1047 sagt, da man zumindest nach DATA *lange* warten sollte,
	 * weil die Nachrichtenverarbeitung durchaus eine Weile dauern kann.
	 * Zitat:
	 *   the increasing round-trip times on the Internet, and that some
	 *   processing after the final dot is required, the timeout for reply to
	 *   the final dot should probably be at least 5 minutes and a timeout of
	 *   10 minutes would not be unreasonable.
	 */

	timeout.tv_sec = 10 * 60;			/* seconds */
	timeout.tv_usec = 0;

	/* `select' returns 0 if timeout, 1 if input available, -1 if error. */
	if (select (FD_SETSIZE, NULL, &fds, NULL, &timeout)!=1)
	{
		*priv->smtprepbuffer='\0';
		return 421;				/* Timeout signalisieren */
	}

	while (cont)
	{
		fgets (priv->smtprepbuffer, sizeof (priv->smtprepbuffer), priv->in);
		if (priv->smtprepbuffer[3] != '-')
			cont = 0;
	}
	return (int) strtol (priv->smtprepbuffer, NULL, 10);
}
#endif

static mio_retcode_t 
smtp_close(mio_t *chan)
{
	struct smtp_private *priv;
	assert(chan);
	priv=(struct smtp_private *) chan->priv_data;
	assert(priv);

#if defined(WANT_SMTP)
	if (!priv->batch)
	{
		/* echtes smtp */
		fprintf (priv->out, "QUIT" CRLF);
		(void) smtp_get_reply(priv);
		fclose(priv->out);
		fclose(priv->in);
	}
	else
#endif
	{
		/* wir schreiben in ein Batchfile ... */
		fputs ("QUIT" CRLF,priv->out);
		fclose(priv->out);
	}
	free(priv->fname);
	free(priv->host);
	free(priv);
	chan->priv_data=NULL;
	mio_free(chan);
	return MIO_OK;
}

#if defined(WANT_SMTP)
static mio_retcode_t
smtp_open(const char *host, const char *mode, mio_t **ret_handle, char *errbuf)
{
	int ret;
	char *errtext=NULL;
	struct smtp_private *priv;
	mio_t *neu;

	*errbuf=0;
	*ret_handle=NULL;

	if (strcmp(mode,"w"))
	{
		sprintf(errbuf,"illegal mode %s for smtp_open",mode);
		log(LOG_ERR,"Illegal mode %s for smtp_open",mode);
		return MIO_ERROR;
	}

	neu=mio_alloc(&mio_smtp);
	if (!neu)
	{
		strcpy(errbuf,_("out of memory"));
		log(LOG_ERR,_("out of memory"));
		return MIO_RETRY;
	}

	priv=malloc(sizeof(struct smtp_private));
	if (priv)
		priv->host=strdup(host);
	if (!priv || !priv->host)
	{
		free(priv);
		free(neu);
		strcpy(errbuf, _("out of memory"));
		log(LOG_ERR, _("out of memory"));
		return MIO_RETRY;
	}

	priv->in=NULL;
	priv->out=NULL;
	priv->fname=NULL;
	priv->batch=FALSE;

	if (!fopen_service ("smtp", host,&priv->in,&priv->out,&errtext))
	{
		free(neu);
		free(priv);
		free(priv->host);
		strcpy(errbuf,errtext ? errtext : _("out of memory"));
		log(LOG_ERR,"%s",errtext ? errtext : _("out of memory"));
		return MIO_RETRY;
	}

	setvbuf (priv->in, NULL, _IOLBF, 0);
	setvbuf (priv->out, NULL, _IOLBF, 0);

	ret = smtp_get_reply (priv);	/* Connection establishment reply */
	if (ret == 521)
	{
		/* RFC 1846:
		 * If an SMTP client encounters a host in an MX record that issues a 
		 * 521 greeting message, it must do one of the following two options:
		 *
		 * a) Attempt to deliver it to a different MX host for that domain.
		 * b) Return the mail with an appropriate non-delivery report.
		 *
		 * wir verwenden Variante b - MX-Lookups sind nicht implementiert.
		 */
		log(LOG_INFO,_("remote host doesn't accept mail"));
		strcpy(errbuf, _("remote host doesn't accept mail"));
		fclose (priv->in);
		fclose (priv->out);
		free(priv->host);
		free(priv);
		free(neu);
		return MIO_ERROR;
	}

	if (ret != 220)
	{
		/* Tja, keine Verbindung ... */
		log(LOG_ERR,_("Not welcome! returncode != 220: %d"),ret);
		sprintf(errbuf,_("Not welcome! returncode != 220: %d"),ret);
		fclose (priv->in);
		fclose (priv->out);
		free(priv->host);
		free(priv);
		free(neu);
		return MIO_RETRY;
	}

	/*
	 * Schade, sendmail mag die naechste Zeile nicht.
	 * fprintf (f, "HELO %s%s" CRLF, opts.pollsite, 
	 *    opts.mausdomain);
	 */
	fprintf (priv->out, "HELO %s" CRLF, opts.mysitedomain);

	ret = smtp_get_reply (priv);	/* HELO reply */
	if (ret >= 300)
	{
		log (LOG_ERR,_("HELO failed: %d"),ret);
		sprintf(errbuf, _("HELO failed: %d"),ret);
		fclose (priv->in);
		fclose (priv->out);
		free(priv->host);
		free(priv);
		free(neu);
		return MIO_RETRY;
	}
	neu->priv_data=priv;
	*ret_handle=neu;
	return MIO_OK;
}
#endif

static mio_retcode_t
bsmtp_open(const char *fname, const char *mode, 
	mio_t **ret_handle, char *errbuf)
{
	struct smtp_private *priv;
	mio_t *neu;

	if (strcmp(mode,"w"))
	{
		sprintf(errbuf,"illegal mode %s for bsmtp_open",mode);
		log(LOG_ERR,"Illegal mode %s for bsmtp_open",mode);
		return MIO_ERROR;
	}

	neu=mio_alloc(&mio_batch_smtp);
	if (!neu)
	{
		sprintf(errbuf,_("out of memory"));
		log(LOG_ERR,_("out of memory"));
		return MIO_RETRY;
	}

	priv=malloc(sizeof(struct smtp_private));
	if (priv)
		priv->fname=strdup(fname);
	if (!priv || !priv->fname)
	{
		if (priv)
			free(priv);
		free(neu);
		sprintf(errbuf, _("out of memory"));
		log(LOG_ERR, _("out of memory"));
		return MIO_RETRY;
	}

	priv->in=NULL;
	priv->batch=TRUE;
	priv->host=NULL;

	priv->out=fopen(fname,"w");
	if (!priv->out)
	{
		int errno1=errno;
		sprintf(errbuf, _("cannot open %s: %s"),fname,strerror(errno1));
		log(LOG_ERR, _("cannot open %s: %s"),fname,strerror(errno1));
		free(priv);
		free(priv->fname);
		free(neu);
		return MIO_ERROR;
	}
	setvbuf (priv->out, NULL, _IOFBF, 0);

	fprintf (priv->out, "HELO %s%s" CRLF, opts.pollsite,  opts.mausdomain);

	neu->priv_data=priv;
	*ret_handle=neu;
	return MIO_OK;
}

#ifdef WANT_SMTP
static mio_retcode_t 
retcode_and_log(struct smtp_private *priv, int r, const char *msgid)
{
	const char *host=priv->host;

	switch(r)
	{
	case 421: /* Service not available, closing transmission channel */
		/* auch timeout! */
		log(LOG_ERR, _("connection lost"));
		return MIO_RETRY;
	case 500: /* Syntax error, command unrecognized */
	case 501: /* Syntax error in parameters or arguments */
	case 502: /* Command not implemented */
	case 503: /* Bad sequence of commands */
	case 504: /* Command parameter not implemented */
		/* Weitermachen sinnlos. */
		log(LOG_ERR,"%s (%s)",format_reply(priv),host);
		return MIO_ERROR;
	case 521: /* host does not accept mail */
		msglog("? %s: SMTP %s (%s)",msgid, format_reply(priv),host);
		return MIO_RETRY; /* Auf einem anderen Weg nochmal versuchen */
	case 551: /* User not local, please try <forward-path> */
		msglog("? %s: SMTP %s (%s)",msgid, format_reply(priv),host);
		return MIO_RETRY; /* Auf einem anderen Weg nochmal versuchen */
	case 552: /* exceeded storage allocation */
	case 553: /* mailbox name not allowed */
	case 554: /* Transaction failed */
	case 450: /* mailbox unavailable / busy */
	case 451: /* local error in processing */
	case 453: /* insufficient system storage */
		msglog("? %s: SMTP %s (%s)",msgid, format_reply(priv),host);
		return MIO_RETRY;
	default: /* Was macht man mit unbekannten Meldungen? */
		msglog("? %s: SMTP %s (%s)",msgid, format_reply(priv),host);
		log(LOG_WARNING,_("unknown reply '%s' (%s)"),format_reply(priv),host);
		if (r>=500)
			return MIO_ERROR;
		return MIO_RETRY;
	}
}
#endif /* WANT_SMTP */

#define GET_REPLY() ((priv->batch) ? 200 : smtp_get_reply(priv))
static mio_retcode_t
smtp_export(mio_t *chan, metamsg_t *meta)
{
#ifdef WANT_SMTP
	int ret;
#endif
	FILE *out;
	FILE *in;
	struct smtp_private *priv;

	assert(chan);
	assert(meta);
	priv=(struct smtp_private *) chan->priv_data;

	in=priv->in;
	out=priv->out;

	if (!out)
	{
		log(LOG_ERR,_("Aieee - no smtp filehandle!"));
		mio_set_errtext(chan, _("no smtp filehandle!"));
		msglog("  %s: no SMTP filehandle (can't happen)",meta->message_id);
		return MIO_RETRY;
	}

	/* SMTP-Envelope */
	fprintf (out, "MAIL FROM:<%s>" CRLF, 
			 meta->env_from ? meta->env_from : "");
#ifdef WANT_SMTP
	ret = GET_REPLY();	/* MAIL reply */
	if (ret > 299)
	{
		return retcode_and_log(priv,ret,meta->message_id);
	}
#endif
	fprintf (out, "RCPT TO:<%s>" CRLF, meta->env_to);
#ifdef WANT_SMTP
	ret = GET_REPLY();	/* MAIL reply */
	if (ret>299)
	{
		return retcode_and_log(priv,ret,meta->message_id);
	}
#endif

	fprintf (out, "DATA" CRLF);
#ifdef WANT_SMTP
	ret = GET_REPLY(); 
	/* Bei irgendeiner Form von Fehler die Nachricht
	 * in einen Mailbatch schreiben. Es gibt nichts, was
	 * sonst getan werden kann.
	 */
	if (ret>399)
	{
		return retcode_and_log(priv,ret,meta->message_id);
	}
#endif
	/* Nachricht schreiben, Zeilentrenner ist CRLF, Punkte escapen */
	if (meta->file) {
		char *lptr=NULL;
		size_t size;
		size_t got;
		while ((got=getdelim(&lptr,&size,'\n',meta->file))!=-1) {
			lptr[got-1]=0;
			if (lptr[got-2]=='\r')
				lptr[got-2]=0;
			if (*lptr=='.')
				fputc('.',out);
			fputs(lptr,out);
			fputs(CRLF,out);
		}
		free(lptr);
	} else
		rfc_print_message(meta,out,CRLF,1);
	fprintf (out, "." CRLF);
#ifdef WANT_SMTP
	ret = GET_REPLY();
	if (ret>299)
	{
		return retcode_and_log(priv,ret,meta->message_id);
	}
	if (!priv->batch)
		msglog("  %s: SMTP sent (%s)",meta->message_id,priv->host);
	else
#endif
		msglog("  %s: SMTP sent (batch)",meta->message_id);
	return MIO_OK;
}
#undef GET_REPLY
