/* Mauscon: Nachrichtenformatskonverter fuer das MausNet
   Copyright (C) 1992 1993 1994 1995 Michael Kraehe, Uwe Ohse

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   rfc2maus.c: Konvertierung von RFC nach MausTausch
*/

#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>

#include <errno.h>
#include <ctype.h>
#include <fcntl.h>
#include <getopt.h>
#include <syslog.h>

#if HAVE_DIRENT_H
# include <dirent.h>
# define NAMLEN(dirent) strlen((dirent)->d_name)
#else
# define dirent direct
# define NAMLEN(dirent) (dirent)->d_namlen
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# if HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# if HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif

#ifdef HAVE_SYS_STAT_H
#  include <sys/stat.h>
#endif

#include "mausconlib.h"
#include "mauscon.h"
#include "metamsg.h"
#include "errormail.h"
#include "rfc.h"
#include "maus.h"
#include "groups.h"
#include "domains.h"
#include "rfcaddr.h"
#include "msg_io.h"
#include "long-options.h"

/* Weil MAUS \r zwingend braucht */
#define LINEEND "\r\n"


int argument_only_convert;
int work_only;
int dont_use_nntp;
int dont_use_smtp;
char *outputfilename;


static void write_cmd(FILE *outputfile);
static BOOLEAN rfc2maus(mio_t *ochan, metamsg_t *meta, char *unlink_fname);
static void convert_news (void);
static void convert_mail (mio_t *ochan);
static void	append_news(FILE *outputfile);


static BOOLEAN
rfc2maus(mio_t *ochan, metamsg_t *msg, char *unlink_fname)
{
	cancel_t *cancel;
	unsigned const char *error_text = NULL;
	int errors_detected;

	errors_detected = 0;
	error_text=NULL;

	if (msg->message_id && !msg->is_mail)
	{
		cancel=cancel_find_in_list(msg->message_id);
		if (cancel 
			&& cancel->from && msg->head_from && msg->head_from->text
			&& !strcasecmp(msg->head_from->text,cancel->from))
		{
			/* Eine Cancel-Nachricht auf die zu verarbeitende Nachricht ist 
			 * also schon eingetroffen. 
			 */		
			msglog("%s: canceled\n",msg->message_id);
			if (unlink_fname)
				unlink(unlink_fname);
			cancel->done=1;
			return TRUE;
		}
	}
	if (!msg->is_mail)
	{
		char *p=NULL;
		BOOLEAN is_control=FALSE;
		if (msg->control)
		{
			p=msg->control;
			is_control=TRUE;
		}
		else if (msg->also_control)
			p=msg->also_control;
		if (msg->supersedes)
			p=msg->supersedes;
		else if (p)
		{
			if (!strncmp(p,"cancel ",7))
				p+=7;
			else
				p=NULL;
		}
		if (p && *p=='<')
		{
			char *q;
			p++;
			q=strchr(p,'>');
			if (q && msg->head_from && msg->head_from->text)
			{
				*q='\0';
				cancel_add(p,msg->head_from->text);
				*q='>';
			}
		}
		if (is_control)
			return TRUE;
	}

	/*
	 * Der Header ist in die Datenstrukturen konvertiert worden.
	 * Nun Vollstndigkeit prfen. Allzu viel ist nicht zu tun.
	 * - bei Mail wird die Angabe der Envelope-Adressen geprft, und
	 *   geprft, ob der Envelope-Absender durch das Gate schreiben darf.
	 * - bei News wird geprft, ob eine der Gruppen bekannt ist, ob
	 *   From oder Sender durch das Gate schreiben darf.
	 * Diese Prfungen mssen wohl beim Boxmodus anders gehandhabt werden
	 * (eine Nachricht, die bis hier kommt, ist als OK anzusehen?)
	 */
	 if (msg->is_mail)
	 {
	 	if (!argument_only_convert)
	 	{

			if (!msg->env_to)
			{
				errors_detected++;
				error_text = "no receiver in envelope";
			}

			if (!msg->env_from)
			{
				errors_detected++;
				error_text = "no sender in envelope";
			}
			else if (!is_allowed(msg->env_from))
			{
				errors_detected++;
				error_text = "envelope-sender is not allowed to use the gateway";
			}
		}
	 }
	 else
	 {
	 	/* News */
	 	if (msg->newsgroups)
	 	{
	 		metalist_t *ng;
	 		BOOLEAN found=FALSE;
			BOOLEAN no_import=FALSE;
	 		ng=msg->newsgroups;
	 		while(ng)
 			{
	 			grp_data_t *grp;
				grp = grp_suche_rfcname (ng->text,opts.pollsite);
				if (grp)
				{
					found=TRUE;
					no_import|=grp->no_import;
					if (grp->mausnet_distribution)
					{
						BOOLEAN do_it=TRUE;
						if (msg->distribution)
						{
							/* Wenn die Distribution nicht schon lokal ist */
							if (strcasecmp(msg->distribution, 
									opts.usenet_dist_lokal))
								free(msg->distribution);
							else
								do_it=FALSE;
						}
						if (do_it)
							msg->distribution=
								XSTRDUP(grp->mausnet_distribution);
					}
				}
 				ng=ng->next;
	 		}
			if (no_import && opts.workmode==GATE)
			{
				errors_detected++;
				error_text = _("import into newsgroup not allowed");
			}
	 		if (!found)
	 		{
				/* 
				 * Dummyeintrag in der Namensliste erzeugen, damit ber
				 * einen Gruppenamen nur einmal gemeckert wird.
				 */
	 			grp_data_t *grp;
				grp=XMALLOC(sizeof(grp_data_t));
				*grp=leer_grp;
				grp->mausname=XSTRDUP(msg->newsgroups->text);
				grp->rfcname=XSTRDUP(msg->newsgroups->text);
				grp->boxen=XSTRDUP(opts.pollsite);
				grp->dont_save=1;
				grp_append_new(grp);
			}
	 	}
	 	else
	 	{
			errors_detected++;
			error_text = _("no newsgroup given");
	 	}

	 	if (msg->tm_date==NULL)
	 	{
			errors_detected++;
			error_text = _("date not given or date unparsable");
	 	}

	 	if (!msg->head_from || !msg->head_from->text)
	 	{
	 		/* Keine From-Zeile? Laut 1036 verboten */
			errors_detected++;
			error_text = _("from not given");
	 	}
	 	else
	 	{
	 		/* news haben laut 1036-Nachfolger nur einen Sender/From */
	 		int ok=0;
	 		if (is_allowed(msg->head_from->text))
	 			ok=1;
	 		else if (msg->sender && msg->sender->text)
	 		{
	 			if (is_allowed(msg->sender->text))
	 				ok=1;
				else
					error_text = _("from and sender address are not allowed to use the gateway");
	 		}
			else
				error_text = _("from address is not allowed to use the gateway");
	 		if (!ok)
	 		{
				errors_detected++;
				if (opts.unlink_unallowed && unlink_fname)
					unlink(unlink_fname);
			}
		}
	}
	if (!msg->message_id)
	{
		if (!msg->is_mail)
		{
			errors_detected++;
			error_text = _("no message-id given");
		}
		else
		{
			char idbuf[256];
			static int counter=0;
			sprintf(idbuf,"No.Id.Given.%lu.%d@%s",
				(unsigned long) getpid(), ++counter,opts.mysitedomain);
			msg->message_id=XSTRDUP(idbuf);
			msglog("[no msgid given -> %s]\n",msg->message_id);
		}
	}
	if (opts.workmode==USER && !msg->is_mail)
	{
		/* Wiedereinleitungen von Nachrichten vermeiden - das *sollte*
		 * zwar auch durch die Absenderkontrolle abgefangen werden,
		 * aber da es so leicht ist, es Newssystem zu verkonfigurieren,
		 * ist diese Paranoia zwecks Dupevermeidung sinnvoll.
		 */
		if (msg->path && strstr(opts.pollsite,msg->path))
		{
			errors_detected++;
			error_text = _("reimport stopped");
		}
	}

	if (errors_detected != 0)
	{
		/* Ob stderr der richtige Platz ist??? */
		if (msg->message_id)
			fprintf (stderr, "ID:%s\n", msg->message_id);

		fprintf (stderr, _("%d errors detected\n"), errors_detected);
		fprintf (stderr, _("one of them: %s\n"), error_text);
		return FALSE;
	}

	/***********************************************************************
	*
	*                    Nun Ausgabe im MausTausch-Format
	*
	************************************************************************/
	ochan->fn->export(ochan,msg);

	return TRUE;
}

static void 
usage(int exitcode, const char *what)
{
    FILE *f=stdout;

    if (exitcode)
    {
        if (what)
            fprintf(stderr, "%s: %s\n",program_name,what);
        fprintf (stderr, _("Try `%s --help' for more information.\n"),
            program_name);
        exit(exitcode);
    }

    fprintf(f, "%s - %s\n", program_name, PACKAGE_VERSION);

    fprintf(f,_("Usage: %s [options] file ...\n"),
        program_name);
    fputs(_("Convert and send MausTausch to RFC\n"),f);
	fputs(_(
"  -c  --cnf-file FILE         use config file FILE\n"
"  -h  --help                  this text\n"
"      --dont-use-nntp         do not use NNTP, just spool into incoming\n"
"      --dont-use-smtp         do not use SMTP, just spool into incoming\n"
"      --only-mail             only mail, no news, no spezial files\n"
"      --only-news             only news, no mail, no spezial files\n"
"      --only-special          only special files, no news, no mail\n"
"  -O, --only-convert          do not send anything, just print to stdout\n"
"  -o, --output FILE           write data to file instead of stdout\n"
"  -v  --version               show version info\n"
"\n"
"short options use the same arguments as the long ones\n"
    ),f);
    exit(exitcode);
}

static void
usage1(int ecode)
{
    usage(ecode,NULL);
}

int
main (int argc, char **argv)
{
	const char *cnffname=SYSCONFDIR "/" CNFNAME;
	mio_t *ochan;
	char errbuf[PATH_MAX];

	struct option cmdlinopts[] =
	{
		{"cnffile", required_argument, NULL, 'c'},
		{"cnf-file", required_argument, NULL, 'c'},
		{"only-convert", no_argument, NULL, 'O'},
		{"dont-use-nntp", no_argument, &dont_use_nntp, 1},
		{"dont-use-smtp", no_argument, &dont_use_smtp, 1},
		{"only-mail", no_argument, &work_only, ONLY_MAIL},
		{"only-news", no_argument, &work_only, ONLY_NEWS},
		{"only-special", no_argument, &work_only, ONLY_SPECIAL},
		{"output", required_argument, NULL, 'o'},
		{NULL, 0, NULL, 0}
	};
	int option_index = 0;
	const char *slash;

	program_name=argv[0];
	slash=strrchr(program_name,'/');
	if (slash)
		program_name=slash+1;

	parse_long_options (argc, argv, program_name, PACKAGE_VERSION, usage1);

    while (1)
    {
        int c;

        c = getopt_long (argc, argv, "c:o:O", cmdlinopts, &option_index);

        /* Detect the end of the options.   */
        if (c == -1)
            break;
        switch (c)
        {
        default:
            break;
        case 'c':
            cnffname = optarg;
            break;
        case 'o':
            outputfilename = optarg;
            break;
        case 'O':
            argument_only_convert = 1;
            break;
        case '?':
            log (LOG_ERR, _("unknown option %s"), argv[optind - 1]);
            usage (2, _("unknown option"));
            exit (2);
            break;
        }
    }                           /* while(1) */

	if (!argv[optind])
	{
		log (LOG_ERR, _("channel not given"));
		exit(2);
	}
	else if (argv[optind+1])
	{
		log (LOG_ERR, _("too many arguments"));
		exit (2);
	}
	action=argv[optind];

	/* configfile lesen */
    if (do_conf (cnffname, argv[optind]))
        exit (1);

	openlog (program_name, 0, opts.sl_facility);
	setlogmask(LOG_UPTO(opts.sl_priority));
	open_msglog();

	if (!grp_load_list(FALSE))
		exit (1);				/* Fehlermeldung schon ausgegeben */

	if (!dom_init())
		exit(1);

	/* Cancel-Liste ffnen und Lock halten */
	if (cancel_init (1))
		exit (1);				/* Fehlermeldung schon ausgegeben */

	if (MIO_OK!=mio_maustausch.open(
		outputfilename ? outputfilename : "-", 
		"w",
		&ochan,
		errbuf))
	{
		fprintf (stderr, "rfc2maus: %s: %s\n", 
			outputfilename ? outputfilename : "stdin",
			 strerror (errno));
		exit (1);
	}
	convert_mail (ochan);
	ochan->fn->close(ochan);

	convert_news ();

	if (outputfilename)
	{
		FILE *outputfile=fopen(outputfilename,"a");
		if (!outputfile)
		{
			fprintf (stderr, _("rfc2maus: cannot fopen %s: %s\n"), 
				outputfilename, strerror (errno));
		}
		append_news(outputfile);
		write_cmd(outputfile);
		fprintf(outputfile,"#"LINEEND);
	}
	else
	{
		append_news(stdout);
		write_cmd(stdout);
		fprintf(stdout,"#"LINEEND);
	}

	/* Alle Cancels lschen - die sind jetzt verarbeitet */
	cancel_end(1,0);

	closelog ();
	close_msglog();
        
#ifdef LEAK_CHECK
	GC_gcollect();
#endif
	exit (0);
}

static void 
write_cmd(FILE *outputfile)
{
	char fname[PATH_MAX];
	char lockfname[PATH_MAX];
	int fatal_err=0;
	char str[1024];
	char *p;
	FILE *f;
	minfo_t *akt;
	cancel_t *cakt;
	BOOLEAN cmd_written=FALSE;

	sprintf(lockfname,"%s/%s/etc/LOCK",opts.spoolbase,action);
	while (lock_create (lockfname, &fatal_err))
	{
		if (fatal_err)
		{
			log(LOG_ERR, _("cannot lock with %s"),lockfname);
			return;
		}
		sleep(5);
	}

	/* 
	 * Eigentlich mte man jetzt die noch brigen Cancels verarbeiten,
	 * sie irgendwie in die MausNetinternen IDs umsetzen ... 
	 * Brute Hack Alert.
	 */
	cakt=cancel_anker();
	if (cakt)
	{
		sprintf(fname,"%s/%s/etc/status.msg",opts.spoolbase,action);
		f=fopen(fname,"a");
		if (f)
		{
			while(cakt)
			{
				if (!cakt->done)
				{
					if (opts.workmode != USER 
						|| strstr(cakt->id,opts.mausdomain))
					{
						fprintf(f,"#%s\n",cakt->id);
						if (cakt->from && opts.boxtyp==BOXTYP_QUARK)
							fprintf(f,"V%s\n",cakt->from);
						fprintf(f,"BX\n");
					}
				}
				cakt=cakt->next;
			}
			fclose(f);
		}
	}

	if (0==info_read_list())
	{
		akt=info_get_anker();
		fprintf(outputfile,"#CMD" LINEEND);
		cmd_written=TRUE;
		while (akt)
		{
			if (akt->needs_crc && akt->subscribed)
			{
				if (time(NULL)>akt->last_date+akt->pause*86400)
					fprintf(outputfile,":%s %ld"LINEEND,akt->kuerzel,
							akt->last_crc);
			}
			else if (akt->subscribed)
			{
				if (time(NULL)>akt->last_date+akt->pause*86400)
					fprintf(outputfile,":%s"LINEEND,akt->kuerzel);
			}
			akt=akt->next;
		}
	}

	sprintf(fname,"%s/%s/etc/infile.cmd",opts.spoolbase,action);
	f=fopen(fname,"r");
	if (f)
	{
		if (!cmd_written)
		{
			fprintf(outputfile,"#CMD" LINEEND);
			cmd_written=TRUE;
		}
		while (fgets (str, sizeof(str), f))
		{
			p=strpbrk(str,"\r\n");
			if (p)
				*p='\0';
			if (*str==':')
				fprintf(outputfile,"%s"LINEEND,str);
			else
				fprintf(outputfile,":%s"LINEEND,str);
		}
		fclose(f);
	}

	sprintf(fname,"%s/%s/etc/once.cmd",opts.spoolbase,action);
	f=fopen(fname,"r");
	if (f)
	{
		if (!cmd_written)
			fprintf(outputfile,"#CMD" LINEEND);
		while (fgets (str, sizeof(str), f))
		{
			p=strpbrk(str,"\r\n");
			if (p)
				*p='\0';
			if (*str==':')
				fprintf(outputfile,"%s"LINEEND,str);
			else
				fprintf(outputfile,":%s"LINEEND,str);
		}
		fclose(f);
		log(LOG_INFO,_("added once.cmd"));
	}
	sprintf(fname,"%s/%s/etc/status.msg",opts.spoolbase,action);
	f=fopen(fname,"r");
	if (f)
	{	
		unlink(fname);
		while (fgets (str, sizeof(str), f))
		{
			p=strpbrk(str,"\r\n");
			if (p)
				*p='\0';
			fprintf(outputfile,"%s"LINEEND,str);
		}
		fclose(f);
	}
	lock_delete(lockfname);
}

static void
append_news (FILE *outputfile)
{
	DIR *dir;
	struct dirent *entry;
	char dirname[PATH_MAX];

	sprintf (dirname, "%s/%s/outgoing", opts.spoolbase, action);
	tzset();
	dir = opendir (dirname);
	if (!dir)
	{
		unsigned char g[256];
		unsigned char m[1024];
		sprintf (m, "%s: %s\n",
				 dirname, strerror (errno));
		sprintf (g, "-%s", opts.gatemaster);

		errormail (NULL, "error creating infile",
				   m, g, NULL);
		return;
	}

	entry = readdir (dir);
	while (entry)
	{
		if (entry->d_name[0] == 'm' && entry->d_name[1]=='t')
		{
			char buf[32768];
			char mausb[PATH_MAX];
			FILE *in;
			char *p;

			sprintf (mausb, "%s/%s/outgoing/%s", opts.spoolbase, action, entry->d_name);
			in=fopen(mausb,"r");
			if (!in)
			{
				log(LOG_NOTICE,_("trouble appending mausbatch %s"), mausb);
			}
			else
			{
				while (fgets(buf,sizeof(buf),in))
				{
					p=strpbrk(buf,"\r\n");
					if (p)
						*p='\0';
					fprintf(outputfile,"%s\r\n",buf);
				}
				fclose(in);
			}
		}
		entry = readdir (dir);
	}
	closedir (dir);
}


static void
convert_mail (mio_t *ochan)
{
	DIR *dir;
	struct dirent *entry;
	char dirname[PATH_MAX];
	buf_t mbuf={NULL,NULL,0};
	metamsg_t *meta;
	meta=meta_alloc();
	XNULLCHECK(meta);

	sprintf (dirname, "%s/%s/outgoing", opts.spoolbase, action);
	tzset();
	dir = opendir (dirname);
	if (!dir)
	{
		unsigned char g[256];
		unsigned char m[1024];
		sprintf (m, "%s: %s\n",
				 dirname, strerror (errno));
		sprintf (g, "-%s", opts.gatemaster);

		errormail (NULL, "error creating infile",
				   m, g, NULL);
		return;
	}

	while ((entry = readdir (dir))!=NULL)
	{
		char fname[PATH_MAX];
		char *transferidp;

		if (strncmp(entry->d_name,"mail",4)!=0)
			continue;

		sprintf (fname, "%s/%s/outgoing/%s", opts.spoolbase, action, 
			entry->d_name);
		transferidp=strrchr(fname,'/')+1;

		/*
		 * Die Nachricht aus der Datei einlesen in den Puffer.
		 */
		mbuf.next=mbuf.start;
		if (rfc_read_message(fname,&mbuf)!=0)
		{
#ifdef HAVE_LSTAT
			/* Pech gehabt, Nachricht nicht lesbar */
			struct stat sb;
			if (lstat(fname,&sb)==0)
			{
				/* wenn symbolischer Link, dann weg damit */
				if ((sb.st_mode & S_IFMT) == S_IFLNK)
					unlink(fname);
			}
#endif
			continue;
		}

		if (rfc_to_meta(&mbuf,meta))
			/* Fehler beim Umwandeln */
			continue;
		meta->is_mail=1;
		meta->transfer_id=XSTRDUP(transferidp);

		rfc2maus (ochan, meta,fname);
	}
#ifndef CLOSEDIR_VOID
	if (closedir(dir))
	{
		fprintf(stderr,_("closedir(%s) failed\n"),dirname);
	}
#else
	closedir (dir);
#endif
	if (mbuf.start)
		free(mbuf.start);
	meta_free(meta);
}

static void
convert_news (void)
{
	DIR *dir;
	struct dirent *entry;
	char dirname[PATH_MAX];
	metamsg_t *meta;
	meta=meta_alloc();
	XNULLCHECK(meta);

	sprintf (dirname, "%s/%s/outgoing", opts.spoolbase, action);
	tzset();
	dir = opendir (dirname);
	if (!dir)
	{
		unsigned char g[256];
		unsigned char m[1024];
		sprintf (m, "%s: %s\n",
				 dirname, strerror (errno));
		sprintf (g, "-%s", opts.gatemaster);

		errormail (NULL, "error creating infile",
				   m, g, NULL);
		return;
	}

	while ((entry = readdir (dir))!=NULL)
	{
		char newsb[PATH_MAX];
		char mausb[PATH_MAX];
		char errbuf[PATH_MAX];
		char *transferidp;
		BOOLEAN is_suck=FALSE;
		mio_t *ichan;
		mio_t *ochan;
		mio_retcode_t ret;
		long problems=0;
		BOOLEAN mausb_open=FALSE;
		BOOLEAN one_ok=FALSE;
		BOOLEAN may_unlink=FALSE;
		/* Nur Files beginnend mit einer Ziffer oder "nb" interessieren
		 * uns hier.
		 */
		if (! ((entry->d_name[0] == 'n' && entry->d_name[1]=='b')
			|| (entry->d_name[0] == 's' && entry->d_name[1]=='k')
			|| isdigit(entry->d_name[0])))
			continue;
		/* alles, was keine Ziffer am Ende hat, ist falsch */
		if (!(isdigit(entry->d_name[strlen(entry->d_name)-1])))
			continue;
		if (entry->d_name[0] == 's' && entry->d_name[1]=='k')
			is_suck=TRUE;

		/* Pfade erzeugen */
		sprintf (newsb, "%s/%s/outgoing/%s", opts.spoolbase, action, 
			entry->d_name);

		strcpy(mausb,newsb);
		transferidp=strrchr(mausb,'/')+1;
		if (isdigit(*transferidp)) {
			memmove(transferidp+2,transferidp,strlen(transferidp)+1);
			may_unlink=TRUE;
		}
		/* "nbxxx" -> "mtxxx", zwecks Zuordnung bei maus2rfc,
		 * bzw "suxxx" -> "mtxxx", suck-format */
		*transferidp='m';
		transferidp[1]='t';

		if (is_suck && MIO_OK!=mio_suck.open(newsb,"r",&ichan,errbuf))
		{
			struct stat st;
			/* dangling symlink? */
			if (-1==stat(newsb,&st) && errno==ENOENT) {
				unlink(newsb);
				continue;
			}
			log(LOG_NOTICE, _("Trouble opening suckbatch %s: %s"), newsb,
				errbuf);
			continue;
		}
		else if (MIO_OK!=mio_rnews.open(newsb,"r",&ichan,errbuf))
		{
			struct stat st;
			/* dangling symlink? */
			if (-1==stat(newsb,&st) && errno==ENOENT) {
				unlink(newsb);
				continue;
			}
			log(LOG_NOTICE,_("Trouble opening rnewsbatch %s: %s"), newsb,
				errbuf);
			continue;
		}


		do 
		{
			ret=ichan->fn->import(ichan,meta);
			if (ret==MIO_EOF)
				break;
			if (ret==MIO_ERROR)
			{
				log(LOG_NOTICE, _("%sbatch %s corrupt: %s"), 
					is_suck ? "suck" : "rnews",
					newsb,
					ichan->errtext);
				problems++;
				break;
			}
			else if (ret==MIO_RETRY)
			{
				log(LOG_NOTICE,_("%sbatch %s: %s"), 
					is_suck ? "suck" : "rnews",
					newsb,
					ichan->errtext);
				problems++;
			}
			else /* MIO_OK */
			{
				if (!mausb_open && MIO_OK!=mio_maustausch.open(mausb,"w",&ochan,errbuf))
				{
					log(LOG_NOTICE,_("Trouble opening mausbatch %s: %s"), 
						newsb, errbuf);
					exit(1);  /* Weitermachen sinnlos. */
				}
				mausb_open=TRUE;
				meta->transfer_id=XSTRDUP(transferidp);
				one_ok|=rfc2maus (ochan, meta, may_unlink ? newsb : NULL);
			}
		} while (TRUE);

		if (problems!=0)
		{
			char probf[PATH_MAX];
			sprintf (probf, "%s/%s/problems/%s", opts.spoolbase, action, 
				entry->d_name);
			rename(newsb,probf);

		}
		if (mausb_open && !one_ok)
			unlink(mausb);
		/* im Usermodus die Nachricht nicht lschen - wir brauchen
		 * sie noch um sie canceln zu knnen wenn opts.cancel_original
		 * TRUE ist 
		 */
		if (ret!=MIO_ERROR && opts.workmode!=USER && opts.cancel_original)
			unlink(newsb);
		ichan->fn->close(ichan);
		ochan->fn->close(ochan);
	}
	closedir (dir);
	meta_free(meta);
}
