/* MausConnect: Nachrichtenformatskonverter fuer das MausNet
   Copyright (C) 1992 1993 1994 1995 Michael Kraehe, Uwe Ohse

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   srfcmail.c: send metamsg as rfc822-mail 
*/

#include "config.h"

#include <sys/types.h>
#ifdef HAVE_SYS_STAT_H
#  include <sys/stat.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <syslog.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <conflib.h>
#include <time.h>

#include "mausconlib.h"
#include "mauscon.h"
#include "metamsg.h"
#include "rfc.h"
#include "srfcmail.h"
#include "msg_io.h"

/*
   Zwei Methoden werden zur Zeit untersttzt, die erste ist "echtes"
   smtp, die zweite smtp via Batchfile. Das erste ist prinzipiell
   schneller und resourcensparender, das zweite problemloser, da
   smtliche Fehler vom rsmtp gehandhabt werden. 
   Mglicherweise gibt's Dupes, aber das sollte nur dann passieren,
   wenn der smtp-Daemon in ungluecklichen Momenten die 
   Verbindung unterbricht.
 */
#define CRLF "\r\n"


#ifdef WANT_SMTP
static mio_t *local_smtp;
static BOOLEAN local_tried;

static mio_t *direct_smtp;
static char direct_site[128];
#endif

static mio_t *batched_smtp;
static long batchcount;

/* 
 * Aufrufen, wenn es gebraucht wird, d.h., wenn der direkte smtp-Versuch
 * gescheitert ist. Falls auch das nicht klappt, weil das Batchfile
 * nicht geffnet werden kann -> Pech, Ende, Aus. Fr den Gatemodus
 * sollte man das noch mal berdenken, aber fr den Usermodus ist es
 * das beste.
 */
static BOOLEAN 
open_batchfile(void)
{
	char fname[PATH_MAX];
	char *pos;
	char errbuf[1024];
	long i;
	struct stat st;
	sprintf (fname, "%s/%s/incoming/mailbatch.", opts.spoolbase, action);
	pos=fname+strlen(fname);
	i=batchcount;
	do
	{
		i++;
		sprintf(pos,"%ld",i);
	} while (stat(fname,&st)==0 && st.st_size!=0);
	batchcount=i;
	if (MIO_OK==mio_batch_smtp.open(fname,"w",&batched_smtp,errbuf))
	{
		msglog("  <generic>: <bsmtp> opened batchfile %s", fname);
	}
	else
	{
		msglog("  <generic>: <bsmtp> cannot open batchfile %s: %s", 
			fname, errbuf);
		log(LOG_ERR,_("cannot open batchfile %s: %s"), 
			fname, errbuf);
		/* Das ist unsere Fallbackmethode - und wenn das schief geht, 
		 * gibt es zur Zeit keine Mglichkeit mehr, Mails zu versenden.
		 * Deshalb geben wir jetzt auf.
		 */
		exit(1);
	}
	return TRUE; /* OK */
}

BOOLEAN srfcmail(BOOLEAN only_convert, metamsg_t *msg)
{
	char errbuf[1024];
	static int atexit_done=0;
	if (!atexit_done) {
		atexit_done=1;
		atexit(srfcmail_end_all);
	}
	if (!msg)
	{
		log(LOG_WARNING, _("srfcmail: no message"));
		/* Wenn man nichts versendet, kann man auch keine Fehler machen */
		return TRUE;
	}
	if (!msg->env_to)
	{
		log(LOG_WARNING, _("srfcmail: kein Empfaenger"));
		return FALSE;
	}

	if (only_convert)
	{
		rfc_print_message(msg,stdout,"\n",0);
		return TRUE;
	}

#ifdef WANT_SMTP
	/* Zuerst der Versuch, die Nachricht direkt zu verschicken, d.h.,
	 * ber einen anderen Rechner / anderen SMTP-Server */
	if (opts.direct_smtp && !dont_use_smtp)
	{
		char *p;
		p=strchr(msg->env_to,'@');
		if (p)
		{
			if (direct_smtp && strcasecmp(p+1,direct_site)!=0)
			{
				direct_smtp->fn->close(direct_smtp);
				direct_smtp=NULL;
			}
			if (!direct_smtp)
			{
				direct_smtp->fn->close(direct_smtp);
				if (MIO_OK!=mio_smtp.open(p+1,"w",&direct_smtp,errbuf))
				{
					msglog("  %s: cannot open %s: %s",
						msg->message_id,p,errbuf);
					log(LOG_ERR,_("%s: cannot open %s: %s"),msg->message_id,
						p, errbuf);
				}
				else
				{
					strcpy(direct_site,p+1);
					msglog("  %s: opened host %s", msg->message_id,p+1);
				}
			}

			if (direct_smtp)
			{
				mio_retcode_t ret;
				ret=direct_smtp->fn->export(direct_smtp,msg);
				if (ret==MIO_OK)
					return TRUE; /* verschickt */
				else if (ret==MIO_ERROR)
				{
					/* Wir wissen hier nicht, ob der Fehler zur Folge
					 * hat, da auch andere Nachrichten abgelehnt werden.
					 * also lassen wir es besser. 
					 */
					msglog("  %s: cannot send to %s: %s",
						msg->message_id,p,direct_smtp->errtext);
					log(LOG_ERR,_("%s: cannot send to %s: %s"),
						msg->message_id, p, direct_smtp->errtext);
					direct_smtp->fn->close(direct_smtp);
					direct_smtp=NULL;
				}
			}
		}
	}

	/* Nun der Versuch, die Nachricht an den Hauptserver zu schicken 
	 * (an die eigene Kiste).
	 */
	if (local_smtp==NULL && !local_tried && !dont_use_smtp)
	{
		local_tried=TRUE;
		if (MIO_OK==mio_smtp.open(opts.mailhost,"w",&local_smtp,errbuf))
			msglog("  %s: opened host %s", msg->message_id,opts.mailhost);
		else
		{
			msglog("  %s: cannot open %s: %s",
				msg->message_id,opts.mailhost,errbuf);
			log(LOG_ERR, _("%s: cannot open %s: %s"),
				msg->message_id, opts.mailhost, errbuf);
			local_smtp=NULL;
		}
	}
	if (local_smtp)
	{
		mio_retcode_t ret;
		ret=local_smtp->fn->export(local_smtp,msg);
		if (ret==MIO_OK)
			return TRUE; /* verschickt */
	}
#endif
	/* Letzter Versuch, das Batchfile */
	if (!batched_smtp)
	{
		if (!open_batchfile())
			return FALSE;
	}
	if (MIO_OK==batched_smtp->fn->export(batched_smtp,msg))
	{
		msglog("  %s: mailbatch %ld",msg->message_id,batchcount);
		return TRUE;
	}
	msglog("# %s: ERROR %s",msg->message_id,batched_smtp->errtext);
	return FALSE;
}

void
srfcmail_end_all(void)
{
#ifdef WANT_SMTP
	if (direct_smtp)
	{
		direct_smtp->fn->close(direct_smtp);
		direct_smtp=NULL;
	}
	if (local_smtp)
	{
		local_smtp->fn->close(local_smtp);
		local_smtp=NULL;
	}
#endif
	if (batched_smtp)
	{
		batched_smtp->fn->close(batched_smtp);
		batched_smtp=NULL;
	}
}
